/*
 * Copyright (c) 2013, Expedia Affiliate Network
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code
 * retain the above copyright notice, these conditions, and the following
 * disclaimer.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are those
 * of the authors and should not be interpreted as representing official policies,
 * either expressed or implied, of the Expedia Affiliate Network or Expedia Inc.
 */
package com.ean.log;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.logging.Log;

/**
 * This class is designed to hijack the LOG variable in the classes in order to enable testing of the logged messages.
 */
public final class LogHijacker implements Log {
    private final List<LogMessage> messages = new LinkedList<>();

    private Log log;

    private boolean debugEnabled = true;

    private boolean errorEnabled = true;

    private boolean fatalEnabled = true;

    private boolean infoEnabled = true;

    private boolean traceEnabled = true;

    private boolean warnEnabled = true;

    /**
     * Sets debugEnabled to the passed value.
     * @param enabled Whether or not to enable debug logging.
     */
    public void setDebugEnabled(final boolean enabled) {
        this.debugEnabled = enabled;
    }

    /**
     * Sets errorEnabled to the passed value.
     * @param enabled Whether or not to enable error logging.
     */
    public void setErrorEnabled(final boolean enabled) {
        this.errorEnabled = enabled;
    }

    /**
     * Sets fatalEnabled to the passed value.
     * @param enabled Whether or not to enable fatal logging.
     */
    public void setFatalEnabled(final boolean enabled) {
        this.fatalEnabled = enabled;
    }

    /**
     * Sets infoEnabled to the passed value.
     * @param enabled Whether or not to enable info logging.
     */
    public void setInfoEnabled(final boolean enabled) {
        this.infoEnabled = enabled;
    }

    /**
     * Sets traceEnabled to the passed value.
     * @param enabled Whether or not to enable trace logging.
     */
    public void setTraceEnabled(final boolean enabled) {
        this.traceEnabled = enabled;
    }

    /**
     * Sets warnEnabled to the passed value.
     * @param enabled Whether or not to enable warn logging.
     */
    public void setWarnEnabled(final boolean enabled) {
        this.warnEnabled = enabled;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isDebugEnabled() {
        return debugEnabled;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isErrorEnabled() {
        return errorEnabled;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isFatalEnabled() {
        return fatalEnabled;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isInfoEnabled() {
        return infoEnabled;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isTraceEnabled() {
        return traceEnabled;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isWarnEnabled() {
        return warnEnabled;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void trace(final Object o) {
        trace(o, null);
        if (log != null && isTraceEnabled()) {
            log.trace(o);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void trace(final Object o, final Throwable throwable) {
        messages.add(new LogMessage("TRACE " + o.toString(), throwable));
        if (log != null && isTraceEnabled()) {
            log.trace(o, throwable);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void debug(final Object o) {
        debug(o, null);
        if (log != null && isDebugEnabled()) {
            log.debug(o);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void debug(final Object o, final Throwable throwable) {
        messages.add(new LogMessage("DEBUG " + o.toString(), throwable));
        if (log != null && isDebugEnabled()) {
            log.debug(o, throwable);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void info(final Object o) {
        info(o, null);
        if (log != null && isInfoEnabled()) {
            log.info(o);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void info(final Object o, final Throwable throwable) {
        messages.add(new LogMessage("INFO " + o.toString(), throwable));
        if (log != null && isInfoEnabled()) {
            log.info(o, throwable);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void warn(final Object o) {
        warn(o, null);
        if (log != null && isWarnEnabled()) {
            log.warn(o);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void warn(final Object o, final Throwable throwable) {
        messages.add(new LogMessage("WARN " + o.toString(), throwable));
        if (log != null && isWarnEnabled()) {
            log.warn(o, throwable);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void error(final Object o) {
        error(o, null);
        if (log != null && isErrorEnabled()) {
            log.error(o);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void error(final Object o, final Throwable throwable) {
        messages.add(new LogMessage("ERROR " + o.toString(), throwable));
        if (log != null && isErrorEnabled()) {
            log.error(o, throwable);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void fatal(final Object o) {
        fatal(o, null);
        if (log != null && isFatalEnabled()) {
            log.fatal(o);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void fatal(final Object o, final Throwable throwable) {
        messages.add(new LogMessage("FATAL " + o.toString(), throwable));
        if (log != null && isFatalEnabled()) {
            log.fatal(o, throwable);
        }
    }

    /**
     * Retrieves the messages that have been logged through this hijacker in the order they were logged.
     * @return The logged messages.
     */
    public List<LogMessage> getMessages() {
        return messages;
    }

    /**
     * Hijacks the "LOG" variable of the specified class. Does NOT pass through the existing LOG variable.
     * @param clazz The class whose "LOG" variable should be hijacked.
     * @return The LogHijacker instance that has been inserted into the "LOG" variable of the hijacked class.
     * @throws NoSuchFieldException If there is no field whose name is "LOG" in clazz
     * @throws IllegalAccessException If the LOG field has already been instantiated, preventing further modification.
     */
    public static LogHijacker hijackLog(final Class clazz) throws NoSuchFieldException, IllegalAccessException {
        return hijackLog(clazz, false);
    }

    /**
     * Hijacks the "LOG" variable of the specified class and optionally passes through log messages to the original
     * LOG variable.
     * @param clazz The class whose "LOG" variable should be hijacked.
     * @param passThroughLog Whether or not to attempt to pass through messages to the original log variable.
     * @return The LogHijacker instance that has been inserted into the "LOG" variable of the hijacked class.
     * @throws NoSuchFieldException If there is no field whose name is "LOG" in clazz
     * @throws IllegalAccessException If the LOG field has already been instantiated, preventing further modification.
     */
    public static LogHijacker hijackLog(final Class clazz, final boolean passThroughLog)
            throws NoSuchFieldException, IllegalAccessException {
        return hijackLog(clazz, passThroughLog, "LOG");
    }

    /**
     * Hijacks the a log variable of the specified class and optionally passes through log messages to the original
     * LOG variable.
     * @param clazz The class whose "LOG" variable should be hijacked.
     * @param passThroughLog Whether or not to attempt to pass through messages to the original log variable.
     * @param logVariable The variable in clazz that is the log you would like to hijack.
     * @return The LogHijacker instance that has been inserted into the "LOG" variable of the hijacked class.
     * @throws NoSuchFieldException If there is no field whose name is "LOG" in clazz
     * @throws IllegalAccessException If the LOG field has already been instantiated, preventing further modification.
     */
    public static LogHijacker hijackLog(final Class clazz, final boolean passThroughLog, final String logVariable)
            throws NoSuchFieldException, IllegalAccessException {
        final Field modifiersField = Field.class.getDeclaredField("modifiers");
        modifiersField.setAccessible(true);
        //make the log not final
        final Field logField = clazz.getDeclaredField(logVariable);
        logField.setAccessible(true);
        modifiersField.setInt(logField, logField.getModifiers() & ~Modifier.FINAL);
        //get the log hijacker
        final LogHijacker hijacker = new LogHijacker();
        if (passThroughLog) {
            hijacker.log = (Log) logField.get(null);
        }
        logField.set(null, hijacker);
        return hijacker;
    }

    /**
     * A container for the messages that get passed to the hijacker.
     */
    public static class LogMessage {
        private final String message;
        private final Throwable throwable;

        /**
         * Constructs a new LogMessage.
         * @param message The message for this log message.
         * @param throwable The throwable for this message.
         */
        public LogMessage(final String message, final Throwable throwable) {
            this.message = message;
            this.throwable = throwable;
        }

        /**
         * Gets the message sent via this log message.
         * @return The message.
         */
        public String getMessage() {
            return message;
        }

        /**
         * Gets the throwable sent via this log message. Can be null.
         * @return The throwable.
         */
        public Throwable getThrowable() {
            return throwable;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object o) {
            return o instanceof LogMessage && this.hashCode() == o.hashCode();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            final int randomPrime = 31;
            int result = getMessage() != null ? getMessage().hashCode() : 0;
            result = randomPrime * result + (getThrowable() != null ? getThrowable().hashCode() : 0);
            return result;
        }
    }
}
